#include "dataarea.hh"

namespace
{
    static const unsigned char Empty = 0;
}

void DataArea::PrepareVec(vec& v, unsigned offset)
{
    if(offset >= v.size())
    {
        unsigned newsize = offset+1;
        if(v.capacity() < newsize)
        {
            const unsigned step = 2048;
            unsigned newcapacity = (newsize + step) &~ (step-1);
            v.reserve(newcapacity);
        }
        v.resize(newsize);
    }
}

DataArea::vec& DataArea::GetRef(unsigned offset, unsigned& base)
{
    map::iterator i = blobs.upper_bound(offset);
    if(i == blobs.begin()) i = blobs.end(); else --i;
    
    if(i != blobs.end())
    {
        unsigned prev_top = i->first + i->second.size();
        if(prev_top >= offset)
        {
            base = i->first;
            return i->second;
        }
    }
    
    /* Create a new blob. */
    base = offset;
    return blobs[offset];
}

void DataArea::WriteByte(unsigned pos, unsigned char byte)
{
    unsigned base;

/*
    for(map::const_iterator i = blobs.begin(); i != blobs.end(); ++i)
        fprintf(stderr, "Blob at $%X is %u bytes\n", i->first, i->second.size());
    
    fprintf(stderr, "GetRef($%X,$%02X) ", pos,byte);
*/
    vec& vector = GetRef(pos, base);
/*
    fprintf(stderr, "-> %u bytes at $%X\n", vector.size(), base);
*/
    unsigned vecpos = pos - base;
    PrepareVec(vector, vecpos);
    vector[vecpos] = byte;
}

unsigned char DataArea::GetByte(unsigned pos) const
{
    map::const_iterator i = blobs.upper_bound(pos);
    if(i == blobs.begin()) i = blobs.end(); else --i;
    
    if(i != blobs.end())
    {
        unsigned prev_top = i->first + i->second.size();
        if(prev_top > pos)
        {
            unsigned base = i->first;
            return i->second[pos - base];
        }
    }
    /* Nothing found. */
    return Empty;
}

unsigned DataArea::GetBase() const
{
    map::const_iterator i = blobs.begin();
    if(i == blobs.end()) return 0;
    return i->first;
}

unsigned DataArea::GetTop() const
{
    map::const_iterator i = blobs.end();
    if(i == blobs.begin()) return 0;
    --i;
    return i->first + i->second.size();
}

const std::vector<unsigned char> DataArea::GetContent() const
{
    std::vector<unsigned char> result(GetSize(), Empty);
    const unsigned base = GetBase();
    for(map::const_iterator i = blobs.begin(); i != blobs.end(); ++i)
    {
        unsigned pos = i->first - base;
        std::copy(i->second.begin(), i->second.end(), result.begin()+pos);
    }
    return result;
}

const std::vector<unsigned char> DataArea::GetContent(unsigned begin, unsigned size) const
{
    std::vector<unsigned char> result(size, Empty);

    for(map::const_iterator i = blobs.begin(); i != blobs.end(); ++i)
    {
        unsigned begin_offset = i->first;
        unsigned count  = i->second.size();
        unsigned target = begin_offset - begin;
        
        if(begin > i->first)
        {
            unsigned decrease = i->first - begin;
            if(decrease >= count) continue;
            count -= decrease;
            begin_offset = begin;
            target = 0;
        }
        if(target >= size) continue;
        
        if(count > size-target)
        {
            count = size-target;
        }
        
        unsigned pos = begin_offset - i->first;

/*
        fprintf(stderr,
            "GetContent($%X,%u): Copying %u bytes from %u (base $%X) to %u (base $%X)\n",
                begin,size,
                count,
                pos, i->first,
                target, begin);
*/

        std::copy(i->second.begin() + pos,
                  i->second.begin() + count,
                  result.begin() + target);
    }
    return result;
}

unsigned DataArea::FindNextBlob(unsigned where, unsigned& length) const
{
    map::const_iterator i = blobs.lower_bound(where);
    if(i == blobs.end()) { length = 0; return 0; }
    length = i->second.size();
    return i->first;
}
